@@ -4,7 +4,7 @@ import contextlib
 import operator
 from colorsys import hsv_to_rgb, rgb_to_hsv
 from io import BytesIO
-from math import ceil, cos, floor, pi, sin, sqrt, tan
+from math import ceil, floor, pi, sqrt, tan
 from xml.etree import ElementTree
 
 from PIL import Image
@@ -55,14 +55,11 @@ def lighten(color):
 
 def draw_page(page, stream):
     """Draw the given PageBox."""
-    bleed = {
-        side: page.style[f'bleed_{side}'].value
-        for side in ('top', 'right', 'bottom', 'left')}
     marks = page.style['marks']
     stacking_context = StackingContext.from_page(page)
     draw_background(
-        stream, stacking_context.box.background, clip_box=False, bleed=bleed,
-        marks=marks)
+        stream, stacking_context.box.background, clip_box=False,
+        bleed=page.bleed, marks=marks)
     draw_background(stream, page.canvas_background, clip_box=False)
     draw_border(stream, page)
     draw_stacking_context(stream, stacking_context)
@@ -249,26 +246,14 @@ def draw_background(stream, bg, clip_box=True, bleed=None, marks=()):
                 stream.set_color_rgb(*bg.color[:3])
                 stream.set_alpha(bg.color.alpha)
                 painting_area = bg.layers[-1].painting_area
-                if painting_area:
-                    if bleed:
-                        # Painting area is the PDF BleedBox
-                        x, y, width, height = painting_area
-                        painting_area = (
-                            x - bleed['left'], y - bleed['top'],
-                            width + bleed['left'] + bleed['right'],
-                            height + bleed['top'] + bleed['bottom'])
-                    stream.rectangle(*painting_area)
-                    stream.clip()
-                    stream.end()
+                stream.rectangle(*painting_area)
+                stream.clip()
+                stream.end()
                 stream.rectangle(*painting_area)
                 stream.fill()
 
         if bleed and marks:
             x, y, width, height = bg.layers[-1].painting_area
-            x -= bleed['left']
-            y -= bleed['top']
-            width += bleed['left'] + bleed['right']
-            height += bleed['top'] + bleed['bottom']
             half_bleed = {key: value * 0.5 for key, value in bleed.items()}
             svg = f'''
               <svg height="{height}" width="{width}"
@@ -741,6 +726,55 @@ def draw_rect_border(stream, box, widths, style, color):
     stream.fill(even_odd=True)
 
 
+def draw_line(stream, x1, y1, x2, y2, thickness, style, color, offset=0):
+    assert x1 == x2 or y1 == y2  # Only works for vertical or horizontal lines
+
+    with stacked(stream):
+        stream.set_color_rgb(*color[:3], stroke=True)
+        stream.set_alpha(color[3], stroke=True)
+
+        if style == 'dashed':
+            stream.set_dash([5 * thickness], offset)
+        elif style == 'dotted':
+            stream.set_dash([thickness], offset)
+
+        if style == 'double':
+            stream.set_line_width(thickness / 3)
+            if x1 == x2:
+                stream.move_to(x1 - thickness / 3, y1)
+                stream.line_to(x2 - thickness / 3, y2)
+                stream.move_to(x1 + thickness / 3, y1)
+                stream.line_to(x2 + thickness / 3, y2)
+            elif y1 == y2:
+                stream.move_to(x1, y1 - thickness / 3)
+                stream.line_to(x2, y2 - thickness / 3)
+                stream.move_to(x1, y1 + thickness / 3)
+                stream.line_to(x2, y2 + thickness / 3)
+        elif style == 'wavy':
+            assert y1 == y2  # Only allowed for text decoration
+            up = 1
+            radius = 0.75 * thickness
+
+            stream.rectangle(x1, y1 - 2 * radius, x2 - x1, 4 * radius)
+            stream.clip()
+            stream.end()
+
+            x = x1 - offset
+            stream.move_to(x, y1)
+            while x < x2:
+                stream.curve_to(
+                    x + radius / 2, y1 + up * radius,
+                    x + 3 * radius / 2, y1 + up * radius,
+                    x + 2 * radius, y1)
+                x += 2 * radius
+                up *= -1
+        else:
+            stream.set_line_width(thickness)
+            stream.move_to(x1, y1)
+            stream.line_to(x2, y2)
+        stream.stroke()
+
+
 def draw_outlines(stream, box):
     width = box.style['outline_width']
     color = get_color(box.style, 'outline_color')
@@ -873,7 +907,7 @@ def draw_collapsed_borders(stream, table):
                 (y + 1, x - 1), (y + 1, x)], vertical=False)
         segments.append((
             score, style, width, color, 'left',
-            (pos_x - width / 2, pos_y1, 0, pos_y2 - pos_y1)))
+            (pos_x, pos_y1, 0, pos_y2 - pos_y1)))
 
     def add_horizontal(x, y):
         if y == 0 and table.skip_cell_border_top:
@@ -896,7 +930,7 @@ def draw_collapsed_borders(stream, table):
             pos_x2 = column_positions[x] + shift_before
         segments.append((
             score, style, width, color, 'top',
-            (pos_x1, pos_y - width / 2, pos_x2 - pos_x1, 0)))
+            (pos_x1, pos_y, pos_x2 - pos_x1, 0)))
 
     for x in range(grid_width):
         add_horizontal(x, 0)
@@ -914,14 +948,10 @@ def draw_collapsed_borders(stream, table):
 
     for segment in segments:
         _, style, width, color, side, border_box = segment
-        if side == 'top':
-            widths = (width, 0, 0, 0)
-        else:
-            widths = (0, 0, 0, width)
         with stacked(stream):
-            clip_border_segment(stream, style, width, side, border_box, widths)
-            draw_rect_border(
-                stream, border_box, widths, style,
+            bx, by, bw, bh = border_box
+            draw_line(
+                stream, bx, by, bx + bw, by + bh, width, style,
                 styled_color(style, color, side))
 
 
@@ -1028,7 +1058,7 @@ def draw_text(stream, textbox, offset_x, text_overflow, block_ellipsis):
     textbox.pango_layout.reactivate(textbox.style)
     stream.begin_text()
     emojis = draw_first_line(
-        stream, textbox, text_overflow, block_ellipsis, x, y)
+        stream, textbox, text_overflow, block_ellipsis, Matrix(d=-1, e=x, f=y))
     stream.end_text()
 
     draw_emojis(stream, textbox.style['font_size'], x, y, emojis)
@@ -1052,8 +1082,7 @@ def draw_emojis(stream, font_size, x, y, emojis):
         stream.pop_state()
 
 
-def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
-                    angle=0):
+def draw_first_line(stream, textbox, text_overflow, block_ellipsis, matrix):
     """Draw the given ``textbox`` line to the document ``stream``."""
     # Don’t draw lines with only invisible characters
     if not textbox.text.strip():
@@ -1107,12 +1136,6 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
 
     utf8_text = textbox.pango_layout.text.encode()
     previous_utf8_position = 0
-
-    matrix = Matrix(1, 0, 0, -1, x, y)
-    if angle:
-        a, c = cos(angle), sin(angle)
-        b, d = -c, a
-        matrix = Matrix(a, b, c, d) @ matrix
     stream.text_matrix(*matrix.values)
     last_font = None
     string = ''
@@ -1156,9 +1179,25 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
             utf8_position = utf8_positions[i]
 
             offset = glyph_info.geometry.x_offset / font_size
-            if offset:
-                string += f'>{-offset}<'
-            string += f'{glyph:02x}' if font.bitmap else f'{glyph:04x}'
+            rise = glyph_info.geometry.y_offset / 1000
+            if rise:
+                if string[-1] == '<':
+                    string = string[:-1]
+                else:
+                    string += '>'
+                stream.show_text(string)
+                stream.set_text_rise(-rise)
+                string = ''
+                if offset:
+                    string = f'{-offset}'
+                string += f'<{glyph:02x}>' if font.bitmap else f'<{glyph:04x}>'
+                stream.show_text(string)
+                stream.set_text_rise(0)
+                string = '<'
+            else:
+                if offset:
+                    string += f'>{-offset}<'
+                string += f'{glyph:02x}' if font.bitmap else f'{glyph:04x}'
 
             # Ink bounding box and logical widths in font
             if glyph not in font.widths:
@@ -1191,7 +1230,16 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
                 hb_data = harfbuzz.hb_blob_get_data(hb_blob, stream.length)
                 if hb_data != ffi.NULL:
                     svg_data = ffi.unpack(hb_data, int(stream.length[0]))
+                    # Do as explained in specification
+                    # https://learn.microsoft.com/typography/opentype/spec/svg
                     tree = ElementTree.fromstring(svg_data)
+                    defs = ElementTree.Element('defs')
+                    for child in list(tree):
+                        defs.append(child)
+                        tree.remove(child)
+                    tree.append(defs)
+                    ElementTree.SubElement(
+                        tree, 'use', attrib={'href': f'#glyph{glyph}'})
                     image = SVGImage(tree, None, None, stream)
                     a = d = font.widths[glyph] / 1000 / font.upem * font_size
                     emojis.append([image, font, a, d, x_advance, 0])
@@ -1230,59 +1278,10 @@ def draw_first_line(stream, textbox, text_overflow, block_ellipsis, x, y,
     return emojis
 
 
-def draw_wave(stream, x, y, width, offset_x, radius):
-    up = 1
-    max_x = x + width
-
-    stream.rectangle(x, y - 2 * radius, width, 4 * radius)
-    stream.clip()
-    stream.end()
-
-    x -= offset_x
-    stream.move_to(x, y)
-
-    while x < max_x:
-        stream.curve_to(
-            x + radius / 2, y + up * radius,
-            x + 3 * radius / 2, y + up * radius,
-            x + 2 * radius, y)
-        x += 2 * radius
-        up *= -1
-
-
 def draw_text_decoration(stream, textbox, offset_x, offset_y, thickness,
                          color):
     """Draw text-decoration of ``textbox`` to a ``document.Stream``."""
-    style = textbox.style['text_decoration_style']
-    with stacked(stream):
-        stream.set_color_rgb(*color[:3], stroke=True)
-        stream.set_alpha(color[3], stroke=True)
-
-        if style == 'dashed':
-            stream.set_dash([5 * thickness], offset_x)
-        elif style == 'dotted':
-            stream.set_dash([thickness], offset_x)
-
-        if style == 'wavy':
-            thickness *= 0.75
-            draw_wave(
-                stream,
-                textbox.position_x, textbox.position_y + offset_y,
-                textbox.width, offset_x, thickness)
-        else:
-            stream.move_to(textbox.position_x, textbox.position_y + offset_y)
-            stream.line_to(
-                textbox.position_x + textbox.width,
-                textbox.position_y + offset_y)
-
-        stream.set_line_width(thickness)
-
-        if style == 'double':
-            delta = 2 * thickness
-            stream.move_to(
-                textbox.position_x, textbox.position_y + offset_y + delta)
-            stream.line_to(
-                textbox.position_x + textbox.width,
-                textbox.position_y + offset_y + delta)
-
-        stream.stroke()
+    draw_line(
+        stream, textbox.position_x, textbox.position_y + offset_y,
+        textbox.position_x + textbox.width, textbox.position_y + offset_y,
+        thickness, textbox.style['text_decoration_style'], color, offset_x)
