@@ -1,12 +1,12 @@
 """Classes for all types of boxes in the CSS formatting structure / box model.
 
-See http://www.w3.org/TR/CSS21/visuren.html
+See https://www.w3.org/TR/CSS21/visuren.html
 
 Names are the same as in CSS 2.1 with the exception of ``TextBox``. In
 WeasyPrint, any text is in a ``TextBox``. What CSS calls anonymous inline boxes
 are text boxes but not all text boxes are anonymous inline boxes.
 
-See http://www.w3.org/TR/CSS21/visuren.html#anonymous
+See https://www.w3.org/TR/CSS21/visuren.html#anonymous
 
 Abstract classes, should not be instantiated:
 
@@ -56,7 +56,7 @@ from ..css import computed_from_cascaded
 class Box:
     """Abstract base class for all boxes."""
     # Definitions for the rules generating anonymous table boxes
-    # http://www.w3.org/TR/CSS21/tables.html#anonymous-boxes
+    # https://www.w3.org/TR/CSS21/tables.html#anonymous-boxes
     proper_table_child = False
     internal_table_or_caption = False
     tabular_container = False
@@ -71,13 +71,11 @@ class Box:
     is_for_root_element = False
     is_column = False
     is_leader = False
-    is_attachment = False
 
     # Other properties
     transformation_matrix = None
     bookmark_label = None
     string_set = None
-    download_name = None
     footnote = None
     cached_counter_values = None
     missing_link = None
@@ -123,7 +121,7 @@ class Box:
 
         """
         # Overridden in ParentBox to also translate children, if any.
-        if dx == 0 and dy == 0:
+        if dx == dy == 0:
             return
         self.position_x += dx
         self.position_y += dy
@@ -190,7 +188,7 @@ class Box:
     def hit_area(self):
         """Return the (x, y, w, h) rectangle where the box is clickable."""
         # "Border area. That's the area that hit-testing is done on."
-        # http://lists.w3.org/Archives/Public/www-style/2012Jun/0318.html
+        # https://lists.w3.org/Archives/Public/www-style/2012Jun/0318.html
         # TODO: manage the border radii, use outer_border_radii instead
         return (self.border_box_x(), self.border_box_y(),
                 self.border_width(), self.border_height())
@@ -222,15 +220,15 @@ class Box:
         height = self.border_height() - bt - bb
 
         # Fix overlapping curves
-        # See http://www.w3.org/TR/css3-background/#corner-overlap
+        # See https://www.w3.org/TR/css-backgrounds-3/#corner-overlap
         ratio = min([1] + [
             extent / sum_radii
-            for extent, sum_radii in [
+            for extent, sum_radii in (
                 (width, tlrx + trrx),
                 (width, blrx + brrx),
                 (height, tlry + blry),
                 (height, trry + brry),
-            ]
+            )
             if sum_radii > 0
         ])
         return (
@@ -291,12 +289,42 @@ class Box:
             self.is_floated() or self.is_absolutely_positioned() or
             self.is_running() or self.is_footnote())
 
+    def is_monolithic(self):
+        """Return whether this box is monolithic."""
+        # https://www.w3.org/TR/css-break-3/#monolithic
+        return (
+            isinstance(self, AtomicInlineLevelBox) or
+            isinstance(self, ReplacedBox) or
+            self.style['overflow'] in ('auto', 'scroll') or
+            (self.style['overflow'] == 'hidden' and
+             self.style['height'] != 'auto'))
+
     # Start and end page values for named pages
 
     def page_values(self):
         """Return start and end page values."""
         return (self.style['page'], self.style['page'])
 
+    # PDF attachments
+
+    def is_attachment(self):
+        """Return whether this link should be stored as a PDF attachment."""
+        from ..html import element_has_link_type
+
+        if self.element is not None and self.element.tag == 'a':
+            return element_has_link_type(self.element, 'attachment')
+        return False
+
+    # Forms
+
+    def is_input(self):
+        """Return whether this box is a form input."""
+        # https://html.spec.whatwg.org/multipage/forms.html#category-submit
+        if self.style['appearance'] == 'auto' and self.element is not None:
+            if self.element.tag in ('button', 'input', 'select', 'textarea'):
+                return not isinstance(self, (LineBox, TextBox))
+        return False
+
 
 class ParentBox(Box):
     """A box that has children."""
@@ -359,13 +387,18 @@ class ParentBox(Box):
 
     def page_values(self):
         start_value, end_value = super().page_values()
-        if self.children:
-            if len(self.children) == 1:
-                page_values = self.children[0].page_values()
+        # TODO: We should find Class A possible page breaks according to
+        # https://drafts.csswg.org/css-page-3/#propdef-page
+        # Keep only children in normal flow for now.
+        children = [
+            child for child in self.children if child.is_in_normal_flow()]
+        if children:
+            if len(children) == 1:
+                page_values = children[0].page_values()
                 start_value = page_values[0] or start_value
                 end_value = page_values[1] or end_value
             else:
-                start_box, end_box = self.children[0], self.children[-1]
+                start_box, end_box = children[0], children[-1]
                 start_value = start_box.page_values()[0] or start_value
                 end_value = end_box.page_values()[1] or end_value
         return start_value, end_value
@@ -454,6 +487,8 @@ class InlineBox(InlineLevelBox, ParentBox):
     inline box.
 
     """
+    link_annotation = None
+
     def hit_area(self):
         """Return the (x, y, w, h) rectangle where the box is clickable."""
         # Use line-height (margin_height) rather than border_height
@@ -536,7 +571,7 @@ class InlineReplacedBox(ReplacedBox, AtomicInlineLevelBox):
 class TableBox(BlockLevelBox, ParentBox):
     """Box for elements with ``display: table``"""
     # Definitions for the rules generating anonymous table boxes
-    # http://www.w3.org/TR/CSS21/tables.html#anonymous-boxes
+    # https://www.w3.org/TR/CSS21/tables.html#anonymous-boxes
     tabular_container = True
 
     def all_children(self):
@@ -581,9 +616,6 @@ class TableColumnGroupBox(ParentBox):
     internal_table_or_caption = True
     proper_parents = (TableBox, InlineTableBox)
 
-    # Default value. May be overriden on instances.
-    span = 1
-
     # Columns groups never have margins or paddings
     margin_top = 0
     margin_bottom = 0
@@ -618,9 +650,6 @@ class TableColumnBox(ParentBox):
     internal_table_or_caption = True
     proper_parents = (TableBox, InlineTableBox, TableColumnGroupBox)
 
-    # Default value. May be overriden on instances.
-    span = 1
-
     # Columns never have margins or paddings
     margin_top = 0
     margin_bottom = 0
@@ -652,9 +681,22 @@ class TableCellBox(BlockContainerBox):
     """Box for elements with ``display: table-cell``"""
     internal_table_or_caption = True
 
-    # Default values. May be overriden on instances.
-    colspan = 1
-    rowspan = 1
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        # HTML 4.01 gives special meaning to colspan=0
+        # https://www.w3.org/TR/html401/struct/tables.html#adef-rowspan
+        # but HTML 5 removed it
+        # https://html.spec.whatwg.org/multipage/tables.html#attr-tdth-colspan
+        # rowspan=0 is still there though.
+        try:
+            self.colspan = max(int(self.element.get('colspan', '').strip()), 1)
+        except (AttributeError, ValueError):
+            self.colspan = 1
+        try:
+            self.rowspan = max(int(self.element.get('rowspan', '').strip()), 0)
+        except (AttributeError, ValueError):
+            self.rowspan = 1
 
 
 class TableCaptionBox(BlockBox):
@@ -680,6 +722,19 @@ class PageBox(ParentBox):
     def __repr__(self):
         return f'<{type(self).__name__} {self.page_type}>'
 
+    @property
+    def bleed(self):
+        return {
+            side: self.style[f'bleed_{side}'].value
+            for side in ('top', 'right', 'bottom', 'left')}
+
+    @property
+    def bleed_area(self):
+        return (
+            -self.bleed['left'], -self.bleed['top'],
+            self.margin_width() + self.bleed['left'] + self.bleed['right'],
+            self.margin_height() + self.bleed['top'] + self.bleed['bottom'])
+
 
 class MarginBox(BlockContainerBox):
     """Box in page margins, as defined in CSS3 Paged Media"""
